import numpy as np

from mercs.utils.inference_tools import (
    dummy_array,
    pad_proba,
    select_nominal,
    select_numeric,
)

INPUTS = "inputs"
COMPUTE = "compute"


# Main algorithm
def inference_algorithm(g, m_list, i_list, c_list, data, nominal_ids):
    """Add inference information to graph g
    The information is added to the graph passed as parameter, no new object is returned

    reminder: node[1] = node attribute id

    Arguments:
        g {[type]} -- graph
        m_list {[type]} -- models
        i_list {[type]} -- imputation nodes
        c_list {[type]} -- composition nodes
        data {[type]} -- test data to predict
        nominal_ids {[type]} -- identifiers of the nominal attributes
    
    Raises:
        ValueError: raised when a node type cannot be recognized

    """

    # Helper functions to check node type
    def _data_node(kind): return kind == "D"
    def _model_node(kind): return kind == "M"
    def _imputation_node(kind): return kind == "I"
    def _composite_node(kind): return kind == "C"

    nb_rows, _ = data.shape

    g_descriptive_ids = list(g.desc_ids)

    if data is not None:
        g.data = data[:, g_descriptive_ids]
    else:
        g.data = None
    for node in g.nodes():
        if _data_node(node[0]):
            in_degree = g.in_degree(node)
            if in_degree == 0:
                input_data_node(g, node, g_descriptive_ids)
            elif in_degree > 0:
                if node[1] in nominal_ids:
                    nominal_data_node(g, node, m_list, c_list)
                else:
                    numeric_data_node(g, node, m_list, c_list)
        elif _model_node(node[0]):
            model_node(g, node, m_list)
        elif _imputation_node(node[0]):
            imputation_node(g, node, i_list, nb_rows)
        elif _composite_node(node[0]):
            composite_node(g, node, c_list)
        else:
            raise ValueError("Did not recognize node kind of {}".format(node))


# Specific nodes:
# 1. input data
# 2. imputation
# 3. numeric data
# 4. nominal data
# 5. model
def input_data_node(g, node, g_desc_ids):
    def f(rel_idx):
        f1 = select_numeric(rel_idx)
        return f1(g.data)

    g.nodes[node][INPUTS] = g_desc_ids.index(node[1])
    g.nodes[node][COMPUTE] = f


def imputation_node(g, node, i_list, nb_rows):
    # Build function
    def f(n):
        return i_list[node[1]].transform(dummy_array(n)).ravel()

    g.nodes[node][INPUTS] = nb_rows
    g.nodes[node][COMPUTE] = f


def numeric_data_node(g, node, m_list, c_list):
    node_parents = _get_parents(g, m_list, c_list, node)

    def f(parents):
        collector = _numeric_inputs(g, parents)
        return np.mean(collector, axis=0)

    g.nodes[node][INPUTS] = node_parents
    g.nodes[node][COMPUTE] = f


def nominal_data_node(g, node, m_list, c_list):
    node_parents = _get_parents(g, m_list, c_list, node, nominal=True)
    classes = np.unique(np.hstack([c for _, c, _, _ in node_parents]))

    def vote(X):
        max_x = np.argmax(X, axis=1)
        return classes.take(max_x, axis=0)

    def F(parents):
        collector = _nominal_inputs(g, parents, classes)
        return np.sum(collector, axis=0)

    def F2(parents):
        return vote(F(parents))

    g.nodes[node]["classes"] = classes
    g.nodes[node][INPUTS] = node_parents
    g.nodes[node]["compute_proba"] = F
    g.nodes[node][COMPUTE] = F2


def model_node(g, node, m_list):
    model_parents = _model_parents(g, node)

    def f(parents):
        X = _model_inputs(g, parents)
        return m_list[node[1]].predict(X)

    g.nodes[node][INPUTS] = model_parents
    g.nodes[node][COMPUTE] = f

    if hasattr(m_list[node[1]], "predict_proba"):
        def f2(parents):
            X = _model_inputs(g, parents)
            return m_list[node[1]].predict_proba(X)

        g.nodes[node]["compute_proba"] = f2


def composite_node(g, node, c_list):
    return model_node(g, node, c_list)


# Helper functions
def compute(g, node, proba=False):
    result_str = "result"
    compute_str = COMPUTE

    if proba:
        result_str += "_proba"
        compute_str += "_proba"

    r = g.nodes[node].get(result_str, None)
    if r is None:
        i = g.nodes[node].get(INPUTS)
        f = g.nodes[node].get(compute_str)
        g.nodes[node][result_str] = f(i)
        return g.nodes[node][result_str]
    else:
        return r


def _nominal_inputs(g, parents, classes):
    # Returns the 'nominal' inputs of the parent nodes
    collector = []
    for rel_idx, parent_classes, n, model_type in parents:
        if len(parent_classes) == len(classes):
            select_nom = select_nominal(rel_idx)
            X = compute(g, n, proba=True)
            selected = select_nom(X, model_type)
            collector.append(selected)
        else:
            prob = pad_proba(parent_classes, classes)
            select_nom = select_nominal(rel_idx)
            X = compute(g, n, proba=True)
            selected = prob(select_nom(X, model_type))
            collector.append(selected)

    return collector


def _numeric_inputs(g, parents):
    # Returns the 'numeric' inputs of the parent nodes
    collector = []
    for rel_idx, n, model_type in parents:
        select_num = select_numeric(rel_idx)
        X = compute(g, n)
        collector.append(select_num(X))
    return collector


def _model_inputs(g, parents):
    # Returns the 'model' inputs of the parent nodes
    collector = [compute(g, n) for n in parents]
    collector = np.stack(collector, axis=1)
    return collector


def _get_parents(g, m_list, c_list, node, nominal=False):
    # Returns the 'nominal' parents of a node
    parents = []
    for kind, predecessor_idx in g.predecessors(node):
        rel_idx = _rel_idx(predecessor_idx, node[1], kind, m_list, c_list)
        model_type = m_list[predecessor_idx].out_kind
        if nominal:
            classes = _classes(
                predecessor_idx,
                rel_idx,
                kind,
                m_list,
                c_list
            )
            parents.append((rel_idx, classes, (kind, predecessor_idx), model_type))
        else:
            parents.append((rel_idx, (kind, predecessor_idx), model_type))

    return parents


def _model_parents(g, node):
    # Returns the 'model' parents of a node
    idxs = {predecessor_idx: (m, predecessor_idx) for m, predecessor_idx in g.predecessors(node)}

    parents = [n for kind, n in sorted(idxs.items())]

    return parents


def _rel_idx(predecessor_idx, node_idx, kind, m_list, c_list):
    # Calculates the relative id of a node with respect to its predecessor, based on node kind
    if kind == "M":
        return m_list[predecessor_idx].targ_ids.index(node_idx)
    elif kind == "C":
        return c_list[predecessor_idx].targ_ids.index(node_idx)
    else:
        return 0


def _classes(predecessor_idx, rel_idx, kind, m_list, c_list):
    # Returns the classes of a model, based on node kind
    if kind == "M":
        return m_list[predecessor_idx].classes_[rel_idx]
    elif kind == "C":
        return c_list[predecessor_idx].classes_[rel_idx]



import numpy as np


def get_i_o(data, desc_ids, targ_ids, filter_nan=True):
    """ Splits data into X and y sets before fitting the model

    Args:
        data: training data
        desc_ids: descriptive attributes
        targ_ids: target attributes
        filter_nan: indicates if NaN values should be filtered

    Returns:
        i: X values to be passed to the model
        o: y values to be passed to the model
    """
    if filter_nan:
        i_o = data[:, desc_ids+targ_ids]

        mask = ~np.any(np.isnan(i_o), axis=1)
        i_o_filtered = i_o[mask]

        # This works because we ordered the columns very specifically.
        i = i_o_filtered[:, :len(desc_ids)]
        o = i_o_filtered[:, len(desc_ids):]

    else:
        i, o = data[:, desc_ids], data[:, targ_ids]
    return i, o



from functools import partial

import networkx as nx
import numpy as np
from dask import delayed

from mercs.composition import o, x

from mercs.utils.inference_tools import (
    pad_proba,
    select_nominal,
    select_numeric,
)
from mercs.utils.inference_tools_legacy import get_ids
from mercs.utils import debug_print

VERBOSITY = 0

"""
This file has two inference algorithms: base and dask.
Base is called 'legacy' in MERCS, maybe it should be deleted.
Dask is an inference algorithm which makes use of the Dask parallel-computing library.
"""


def base_inference_algorithm(g):
    # Convert the graph to its functions
    sorted_nodes = list(nx.topological_sort(g))

    msg = """
    sorted_nodes:    {}
    """.format(
        sorted_nodes
    )
    debug_print(msg, level=1, V=VERBOSITY)
    functions = {}
    q_desc_ids = list(get_ids(g, kind="desc"))

    for node_name in sorted_nodes:
        node = g.nodes(data=True)[node_name]

        if node.get("kind", None) == "data":
            if len(nx.ancestors(g, node_name)) == 0:
                functions[node_name] = select_numeric(q_desc_ids.index(node["idx"]))
            else:
                # Select the relevant output
                previous_node = [t[0] for t in g.in_edges(node_name)][0]
                previous_t_idx = g.nodes[previous_node]["tgt"]
                relevant_idx = previous_t_idx.index(node["idx"])

                functions[node_name] = o(
                    select_numeric(relevant_idx), functions[previous_node]
                )

        elif node.get("kind", None) == "imputation":
            functions[node_name] = node["function"]

        elif node.get("kind", None) == "model":
            previous_nodes = [t[0] for t in g.in_edges(node_name)]
            inputs = {g.nodes[n]["tgt"][0]: functions[n] for n in previous_nodes}
            inputs = [
                inputs[k] for k in sorted(inputs)
            ]  # We need to sort to get the inputs in the correct order.

            inputs = o(np.transpose, x(*inputs, return_type=np.array))

            f = node["function"]
            functions[node_name] = o(f, inputs)

        elif node.get("kind", None) == "prob":
            # Select the relevant output
            prob_idx = node["idx"]
            prob_classes = node["classes"]

            previous_nodes = [t[0] for t in g.in_edges(node_name)]
            previous_classes = [g.edges[t]["classes"] for t in g.in_edges(node_name)]
            previous_t_idx = [g.nodes[n]["tgt"] for n in previous_nodes]

            inputs = [
                (functions[n], t, c)
                for n, t, c in zip(previous_nodes, previous_t_idx, previous_classes)
            ]

            for idx, (f1, t, c) in enumerate(inputs):
                f2 = o(select_nominal(t.index(prob_idx)), f1)

                if len(c) < len(prob_classes):
                    f2 = o(pad_proba(c, prob_classes), f2)

                inputs[idx] = f2

            f = partial(np.sum, axis=0)
            functions[node_name] = o(f, x(*inputs, return_type=np.array))

        elif node.get("kind", None) == "vote":
            # Convert probabilistic votes to single prediction
            previous_node = [t[0] for t in g.in_edges(node_name)][0]
            functions[node_name] = o(node["function"], functions[previous_node])

        elif node.get("kind", None) == "merge":
            merge_idx = node["idx"]
            previous_nodes = [t[0] for t in g.in_edges(node_name)]
            previous_t_idx = [g.nodes[n]["tgt"] for n in previous_nodes]

            inputs = [(functions[n], t) for n, t in zip(previous_nodes, previous_t_idx)]

            inputs = [
                o(select_numeric(t_idx.index(merge_idx)), f) for f, t_idx in inputs
            ]
            inputs = o(np.transpose, x(*inputs, return_type=np.array))

            f = partial(np.mean, axis=1)
            functions[node_name] = o(f, inputs)

    return functions


def dask_inference_algorithm(g, X=None, sorted_nodes=None):
    if sorted_nodes is None:
        sorted_nodes = list(nx.topological_sort(g))

    functions = {}

    q_desc_ids = list(get_ids(g, kind="desc"))

    if X is None:
        data = None
    else:
        data = delayed(X[:, q_desc_ids])

    for node_name in sorted_nodes:
        kind = g.nodes[node_name]['kind']
        node = g.nodes[node_name]

        if kind in {'imputation'}:
            actions[kind](node, data)
        elif kind in {'data'}:
            actions[kind](g, node, node_name, data, q_desc_ids)
            functions[node_name] = node["dask"]
        elif kind in {'prob'}:
            actions[kind](g, node, node_name)
            functions[node_name] = node["dask"]
        else:
            actions[kind](g, node, node_name)

    return functions


def dask_imputation_node(node, data):
    node["dask"] = delayed(node["function"])(data)
    return


def dask_data_node(g, node, node_name, data, q_desc_ids):
    n_parents = len(g.in_edges(node_name))

    if n_parents == 0:
        idx = node["idx"]
        node["dask"] = delayed(select_numeric(q_desc_ids.index(idx)))(data)
    else:
        # Select the relevant output
        parent_relative_idx, parent_function = _get_parents_of_data_node(g, node, node_name)
        node["dask"] = delayed(select_numeric(parent_relative_idx))(parent_function)
    return


def dask_model_node(g, node, node_name):
    # Collect input data
    parent_functions = _get_parents_of_model_node(g, node)
    collector = delayed(np.stack)(parent_functions, axis=1)

    # Convert function
    node["dask"] = delayed(node["predict"])(collector)

    if "predict_proba" in node:
        node["dask_proba"] = delayed(node["predict_proba"])(collector)

    return


def dask_prob_node(g, node, node_name):
    # Parent nodes
    parent_nodes = [s for s, t in g.in_edges(node_name)]

    parent_functions = [g.nodes[n]["dask_proba"] for n in parent_nodes]
    parent_targets = [g.nodes[n]["tgt"] for n in parent_nodes]
    parent_classes = [g.edges[e]["classes"] for e in g.in_edges(node_name)]

    inputs = zip(parent_functions, parent_targets, parent_classes)

    # Incorporate extra step(s)
    for idx, (f1, t, c) in enumerate(inputs):
        f2 = delayed(select_nominal(t.index(node["idx"])))(f1)

        if len(c) < len(node["classes"]):
            f3 = delayed(pad_proba(c, node["classes"]))(f2)
        else:
            f3 = f2

        # Overwrite parent functions
        parent_functions[idx] = f3

    # Collect everything in one single array
    node["dask"] = delayed(partial(np.sum, axis=0))(parent_functions)
    return


def dask_vote_node(g, node, node_name):
    parent_node = [s for s, t in g.in_edges(node_name)].pop()
    parent_function = g.nodes[parent_node]["dask"]
    classes = node["classes"]

    # Build function
    def vote(X):
        return classes.take(np.argmax(X, axis=1), axis=0)

    node["dask"] = delayed(vote)(parent_function)
    return


def dask_merge_node(g, node, node_name):
    # Parent nodes
    parent_nodes = [s for s, t in g.in_edges(node_name)]

    parent_functions = [g.nodes[n]["dask"] for n in parent_nodes]
    parent_targets = [g.nodes[n]["tgt"] for n in parent_nodes]

    inputs = zip(parent_functions, parent_targets)

    # Incorporate extra step(s)
    for idx, (f1, t) in enumerate(inputs):
        f2 = delayed(select_numeric(t.index(node["idx"])))(f1)
        parent_functions[idx] = f2

    node["dask"] = delayed(partial(np.mean, axis=0))(parent_functions)
    return


actions = dict(
    data=dask_data_node,
    prob=dask_prob_node,
    model=dask_model_node,
    vote=dask_vote_node,
    merge=dask_merge_node,
    imputation=dask_imputation_node,
)


def _get_parents_of_data_node(g, node, node_name):
    # It can only be one parent
    parent_node = g.nodes[[s for s, t in g.in_edges(node_name)].pop()]

    parent_relative_idx = parent_node["tgt"].index(node["idx"])
    parent_function = parent_node["dask"]

    return parent_relative_idx, parent_function


def _get_parents_of_model_node(g, node_name):
    parent_nodes = [s for s, t in g.in_edges(node_name)]
    parent_indices = [g.nodes[n]["idx"] for n in parent_nodes]

    parent_functions = {idx: g.nodes[n]["dask"] for idx, n in zip(parent_indices, parent_nodes)}
    parent_functions = [parent_functions[k] for k in sorted(parent_functions)]

    return parent_functions


def _get_parents_of_prob_node(g, node_name):
    parent_nodes = [s for s, t in g.in_edges(node_name)]
    parent_functions = [g.nodes[n]["dask"] for n in parent_nodes]

    return parent_functions



from functools import partial
import numpy as np

from sklearn.preprocessing import normalize
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor


class DecoratedRandomForestRegressor(RandomForestRegressor):
    def fit(self, X, y, **kwargs):
        super(DecoratedRandomForestRegressor, self).fit(X, y, **kwargs)

        decorate_forest(self)
        return


class DecoratedRandomForestClassifier(RandomForestClassifier):
    def fit(self, X, y, **kwargs):
        super(DecoratedRandomForestClassifier, self).fit(X, y, **kwargs)

        decorate_forest(self)
        return


class DecoratedDecisionTreeClassifier(DecisionTreeClassifier):
    def fit(self, X, y, **kwargs):
        super(DecoratedDecisionTreeClassifier, self).fit(X, y, **kwargs)

        decorate_tree(self)
        return


class DecoratedDecisionTreeRegressor(DecisionTreeRegressor):
    def fit(self, X, y, **kwargs):
        super(DecoratedDecisionTreeRegressor, self).fit(X, y, **kwargs)

        decorate_tree(self)
        return


class Node(object):
    def __init__(
        self,
        idx=0,
        feature=-1,
        threshold=0,
        left_child=-1,
        right_child=-1,
        value=None,
        n_samples=-1,
    ):
        self.idx = idx
        self.feature = feature
        self.threshold = threshold
        self.left_child = left_child
        self.right_child = right_child
        self.value = value
        self.leaf = left_child == right_child
        self.n_samples = n_samples
        return

    def __str__(self):
        msg = "Node {0}. Rule: X[:, {1}] <= {2:.2f}. Leaf: {3}".format(
            self.idx, self.feature, self.threshold, self.leaf
        )
        return msg


def get_nodes(tree):
    n_nodes = tree.node_count
    children_left = tree.children_left
    children_right = tree.children_right
    feature = tree.feature
    threshold = tree.threshold
    value = tree.value
    n_node_samples = tree.n_node_samples

    nodes = []

    for n_idx in range(n_nodes):
        node = Node(
            idx=n_idx,
            feature=feature[n_idx],
            threshold=threshold[n_idx],
            left_child=children_left[n_idx],
            right_child=children_right[n_idx],
            value=value[n_idx],
            n_samples=n_node_samples[n_idx],
        )

        nodes.append(node)

    return nodes


def add_nodes(clf):
    clf.nodes = get_nodes(clf.tree_)
    return


def compute_average(left_value, right_value, left_samples, right_samples):
    nominator = (left_samples * left_value + right_samples * right_value)
    denominator = left_samples + right_samples
    return  nominator/denominator


def apply_tree_on_sample(sample, nodes, n_idx=0):

    node = nodes[n_idx]

    if node.leaf:
        return node.value
    else:
        if sample[node.feature] <= node.threshold:
            return apply_tree_on_sample(sample, nodes, n_idx=node.left_child)
        elif sample[node.feature] > node.threshold:
            return apply_tree_on_sample(sample, nodes, n_idx=node.right_child)
        else:
            left_samples = nodes[node.left_child].n_samples
            right_samples = nodes[node.right_child].n_samples
            left_value = apply_tree_on_sample(sample, nodes, n_idx=node.left_child)
            right_value = apply_tree_on_sample(sample, nodes, n_idx=node.right_child)
            return compute_average(left_value, right_value, left_samples, right_samples)


def apply(X, nodes):
    out = np.zeros((X.shape[0], nodes[0].value.shape[1]))
    for i in range(X.shape[0]):
        out[i, :] = apply_tree_on_sample(X[i, :], nodes)
    return out


def predict(X, clf=None):
    if isinstance(clf, DecisionTreeClassifier):
        return predict_clf(clf, X)
    else:
        return predict_rgr(clf, X)


def predict_clf(clf, X):
    out = predict_proba(X, clf=clf, normalize_out=False)
    return clf.classes_.take(np.argmax(out, axis=1), axis=0)


def predict_proba(X, clf=None, normalize_out=True):

    if hasattr(clf, "nodes"):
        out = apply(X, clf.nodes)
    else:
        add_nodes(clf)
        out = apply(X, clf.nodes)

    if normalize_out:
        normalize(out, norm="l1", copy=False)
    return out


def predict_rgr(clf, X):
    if hasattr(clf, "nodes"):
        out = apply(X, clf.nodes)
    else:
        add_nodes(clf)
        out = apply(X, clf.nodes)
    return out


def fancy_predict(X, tree=None, **kwargs):
    try:
        y_pred = tree.classic_predict(X, **kwargs)
        return y_pred
    except ValueError:
        return predict(X, clf=tree)


def fancy_predict_proba(X, tree=None, **kwargs):
    try:
        y_proba = tree.classic_predict_proba(X, **kwargs)
        return y_proba
    except ValueError:
        return predict_proba(X, clf=tree)


def _validate_X_predict_overwrite(X):
    return X.astype(np.float32)


def decorate_tree(tree):
    add_nodes(tree)
    tree.classic_predict = tree.predict
    tree.predict = partial(fancy_predict, tree=tree)

    if isinstance(tree, DecisionTreeClassifier):
        tree.classic_predict_proba = tree.predict_proba
        tree.predict_proba = partial(fancy_predict_proba, tree=tree)

    return


def decorate_forest(forest):
    for dt in forest:
        decorate_tree(dt)

    forest._validate_X_predict = _validate_X_predict_overwrite

    return




import itertools
import warnings

import numpy as np
import shap
from joblib import Parallel, delayed
from sklearn.preprocessing import normalize

try:
    from lightgbm import LGBMClassifier as LGBMC
    from lightgbm import LGBMRegressor as LGBMR
except:
    LGBMC, LGBMR = None, None

try:
    from catboost import CatBoostClassifier as CBC
    from catboost import CatBoostRegressor as CBR
except:
    CBC, CBR = None, None

try:
    from wekalearn import RandomForestClassifier as WLC
    from wekalearn import RandomForestRegressor as WLR
except:
    WLC, WLR = None, None

from mercs.composition.CanonicalModel import CanonicalModel
from mercs.utils import code_to_query, get_i_o


def base_induction_algorithm(
        data,
        m_codes,
        metadata,
        classifier,
        regressor,
        mixed,
        classifier_kwargs,
        regressor_kwargs,
        mixed_kwargs,
        random_state=997,
        calculation_method_feature_importances="default",
        min_nb_samples=10,
        n_jobs=1,
        verbose=0,
):
    """Basic induction algorithm. Models according to the m_codes it receives.
    
    Arguments:
        data {np.ndarray} -- Input data
        m_codes {np.ndarray} -- Model codes
        metadata {dict} -- Metadata of MERCS
        classifier {Supported ML learner} -- Supported ML learner
        regressor {Supported ML learner} -- Supported ML learner
        mixed {Supported ML mixed learner} -- Supported ML mixed learner
        classifier_kwargs {dict} -- Kwargs for classifier
        regressor_kwargs {dict} -- Kwargs for regressor
        mixed_kwargs {dict} -- kwargs for mixed learner
    
    Keyword Arguments:
        random_state {int} -- Seed for random numbers (default: {997})
        n_jobs {int} -- Joblib can be used in training. (default: {1})
        verbose {int} -- Verbosity level for Joblib. (default: {0})
    
    Raises:
        ValueError: When trying to learn a model with both nominal and numeric outputs.
    
    Returns:
        m_list -- List of learned ML-models
    """

    assert isinstance(data, np.ndarray)

    # Init
    if classifier_kwargs is None:
        classifier_kwargs = dict()
    if regressor_kwargs is None:
        regressor_kwargs = dict()

    n_rows, n_cols = data.shape
    attributes = set(range(n_cols))
    nominal_attributes = metadata.get("nominal_attributes")
    numeric_attributes = metadata.get("numeric_attributes")
    msg = """Not all attributes of m_codes are accounted for in metadata"""
    assert not (attributes - nominal_attributes - numeric_attributes), msg

    # Codes to queries
    ids = [
        (d, t)
        for d, t, _ in [code_to_query(m_code, return_list=True) for m_code in m_codes]
    ]

    # Generate a list of random seeds.
    np.random.seed(random_state)
    random_states = np.random.randint(10 ** 3, size=(len(ids)), dtype=int)

    parameters = _build_parameters(
        ids,
        nominal_attributes,
        classifier,
        classifier_kwargs,
        numeric_attributes,
        regressor,
        regressor_kwargs,
        mixed,
        mixed_kwargs,
        random_states,
        calculation_method_feature_importances,
        min_nb_samples,
        data,
    )

    # Learn Models
    m_list = _build_models(parameters, n_jobs=n_jobs, verbose=verbose)

    return m_list


def expand_induction_algorithm(
        data,
        m_codes,
        metadata,
        classifier,
        regressor,
        classifier_kwargs,
        regressor_kwargs,
        random_state=997,
        n_jobs=1,
        verbose=0,
):
    """Basic induction algorithm. Models according to the m_codes it receives.

    Arguments:
        data {np.ndarray} -- Input data
        m_codes {np.ndarray} -- Model codes
        metadata {dict} -- Metadata of MERCS
        classifier {Supported ML learner} -- Supported ML learner
        regressor {Supported ML learner} -- Supported ML learner
        classifier_kwargs {dict} -- Kwargs for classifier
        regressor_kwargs {dict} -- Kwargs for regressor

    Keyword Arguments:
        random_state {int} -- Seed for random numbers (default: {997})
        n_jobs {int} -- Joblib can be used in training. (default: {1})
        verbose {int} -- Verbosity level for Joblib. (default: {0})

    Raises:
        ValueError: When trying to learn a model with both nominal and numeric outputs.

    Returns:
        m_list -- List of learned ML-models
    """
    m_list = base_induction_algorithm(
        data,
        m_codes,
        metadata,
        classifier,
        regressor,
        None,
        classifier_kwargs,
        regressor_kwargs,
        None,
        random_state=random_state,
        n_jobs=n_jobs,
        verbose=verbose,
    )

    m_list = _expand_m_list(m_list)

    return m_list


def _expand_m_list(m_list):
    return list(itertools.chain.from_iterable(m_list))


def _build_models(parameters, n_jobs=1, verbose=0):
    """ Method in charge of learning the models based on the given parameters. It can be done is parallel by specifying
    the number of jobs.

    Args:
        parameters: configuration parameters of the models
        n_jobs: number of parallel jobs
        verbose: verbosity level for multi-core learning

    Returns:
        m_list: list of trained models

    """

    if n_jobs < 2:
        if n_jobs < 1:
            msg = """Number of jobs needs to be at least 1. Assuming 1 job."""
            warnings.warn(msg)
        m_list = [_learn_model(*a, **k) for a, k in parameters]
    else:
        msg = """Training is being parallellized using Joblib. Number of jobs = {}""".format(
            n_jobs
        )
        warnings.warn(msg)

        m_list = Parallel(n_jobs=n_jobs, verbose=verbose)(
            delayed(_learn_model)(*a, **k) for a, k in parameters
        )
    return m_list


def _build_parameters(
        ids,
        nominal_attributes,
        classifier,
        classifier_kwargs,
        numeric_attributes,
        regressor,
        regressor_kwargs,
        mixed,
        mixed_kwargs,
        random_states,
        calculation_method_feature_importances,
        min_nb_samples,
        data,
):
    """ Creates the "parameters" object used by the models

    Args:
        ids: query ids in the form of [(descriptive, target), (descriptive, target), ...]
        nominal_attributes: idxs of the nominal attributes
        classifier: classifier algorithm
        classifier_kwargs: classifier configuration
        numeric_attributes: idxs of the numeric attributes
        regressor: regressor algorithm
        regressor_kwargs: regressor configuration
        mixed: mixed learning algorithm
        mixed_kwargs: mixed learner configuration
        random_states:
        calculation_method_feature_importances:
        min_nb_samples: minimum number of samples
        data: training data

    Returns:
        parameters: list of the model parameters
            - data: training data
            - desc_ids: descriptive parameters
            - targ_ids: target parameters
            - learner: learner algorithm
            - out_kind: output data type(nominal, numeric, mixed)
            - kwargs: model kwargs

    """
    parameters = []
    for idx, (desc_ids, targ_ids) in enumerate(ids):
        if set(targ_ids).issubset(nominal_attributes):
            learner = classifier
            out_kind = "nominal"
            kwargs = classifier_kwargs.copy()  # Copy is essential
            kwargs["random_state"] = random_states[idx]
        elif set(targ_ids).issubset(numeric_attributes):
            learner = regressor
            out_kind = "numeric"
            kwargs = regressor_kwargs.copy()  # Copy is essential
            kwargs["random_state"] = random_states[idx]
        else:
            # Case when target ids contain both numerical and nominal data
            learner = mixed
            out_kind = "mixed"
            kwargs = mixed_kwargs.copy()
            kwargs["classification_targets"] = np.where(
                np.array(list(nominal_attributes)) == np.array(targ_ids)
            )[0].tolist()

        kwargs["calculation_method_feature_importances"] = calculation_method_feature_importances
        kwargs["min_nb_samples"] = min_nb_samples

        kwargs = _add_categorical_features_to_kwargs(
            learner, desc_ids, nominal_attributes, kwargs
        )  # This is learner-specific

        # Learn a model for current desc_ids-targ_ids combo
        args = (data, desc_ids, targ_ids, learner, out_kind)
        parameters.append((args, kwargs))

    return parameters


def _learn_model(
        data,
        desc_ids,
        targ_ids,
        learner,
        out_kind,
        filter_nan=True,
        min_nb_samples=10,
        calculation_method_feature_importances="default",
        **kwargs
):
    """Learn a single model from the data.

    The arguments of this function determine specifics on which task,
    which learner etc.

    Model is a machine learning method that has a .fit() method.

    Args:
        data: training data
        desc_ids: ids of the descriptive attributes
        targ_ids: ids of the target attributes
        learner: learning algorithm
        out_kind: type of the ouput data (numeric, nominal or mixed)
        filter_nan: indicates if NaN values should be filtered
        min_nb_samples: minimum number of samples
        calculation_method_feature_importances:
        **kwargs: keyword argument

    Returns:
        model: the learned model
    """
    assert learner is not None

    i, o = get_i_o(data, desc_ids, targ_ids, filter_nan=filter_nan)

    if i.shape[0] < min_nb_samples:
        msg = """
        Only {} samples available for training.
        min_nb_samples is set to {}.
        Therefore no training occured.
        """.format(
            i.shape[0], min_nb_samples
        )
        warnings.warn(msg)
        return None
    else:
        # Pre-processing
        if i.ndim == 1:
            # We always want 2D inputs
            i = i.reshape(-1, 1)

        multi_target = o.shape[1] != 1
        if not multi_target:
            # If output is single variable, we need 1D matrix
            o = o.ravel()

        if learner in {LGBMC, LGBMR}:
            categorical_feature = kwargs.pop("categorical_feature")
            model = learner(**kwargs)
            model.fit(i, o, categorical_feature=categorical_feature)
        else:
            model = learner(**kwargs)
            model.fit(i, o)

        performance = 1.0

        if calculation_method_feature_importances in {"shap"}:
            model.shap_values_ = _calculate_shap_values(model, i)

        # Bookkeeping
        model = CanonicalModel(model, desc_ids, targ_ids, out_kind, performance)

        return model


# Helpers
def _calculate_shap_values(model, X):
    shap_values = shap.TreeExplainer(model).shap_values(X)
    if isinstance(shap_values, list):
        r = _summarize_shaps(shap_values[0])
    else:
        r = _summarize_shaps(shap_values)
    return r


def _summarize_shaps(shap_values):
    avgs_values = np.mean(np.abs(shap_values), axis=0)
    return np.squeeze(normalize(avgs_values.reshape(1, -1), norm="l1"))


def _add_categorical_features_to_kwargs(learner, desc_ids, nominal_attributes, kwargs):
    assert learner is not None
    cat_features = _get_cat_features(desc_ids, nominal_attributes)

    if learner in {CBC, CBR}:
        kwargs["cat_features"] = cat_features
    elif learner in {WLC, WLR}:
        kwargs["cat_features"] = cat_features
    elif learner in {LGBMC, LGBMR}:
        kwargs["categorical_feature"] = cat_features

    return kwargs


def _get_cat_features(desc_ids, nominal_ids):
    cat_features = np.where(
        np.in1d(np.array(list(desc_ids)), np.array(list(nominal_ids)))
    )[0]
    return [int(c) for c in cat_features]
